/*
 * oh how perl/c++ creeps into my c
 * an array object :)
 *
 * $Id: array_t.c,v 1.5 2001/10/04 20:25:26 estabroo Exp $
 *
 */

#include <stddef.h>
#include <sys/types.h>
#include <syslog.h>
#include <string.h>
#include <stdlib.h>

#include "array_t.h"

/*
  might want to have a dynamic down realloc to help with memory for
   arrays that have been allocated huge amounts of space, but are
   now only using a small portion of it
*/

inline static char* _strdup(char* s) {
  if (s == NULL) {
    return NULL;
  }
  return strdup(s);
}

/* allocate memory for the array
    0 it's big enough already
    > 0 new size
    < 0 memory problem when allocating
*/

inline static int _alloc_array(array_t* self, int size) {
  char* ptr;
  int   i;

  if (self == NULL) { /* no such array */
    return -1;
  }

  if (self->_n_alloc > size) {
    return 0;
  }
  i = ((size/16)+1)*16; /* make it a mulituple of 16 */
  if (self->_n_alloc == 0) {
    self->_p = malloc(i*sizeof(char*));
    if (self->_p == NULL) {
      syslog(LOG_CRIT, "memory error allocating %d elements : %m", i);
      return -1;
    }
  } else {
    ptr = realloc(self->_p, i*sizeof(char*));
    if (ptr == NULL) {
      syslog(LOG_CRIT, "memory error allocating %d elements : %m", i);
      return -1;
    }
  }
  memset((void*)&(self->_p[self->_n_alloc]), 0, (i - self->_n_alloc)*sizeof(char*));
  self->_n_alloc = i;
  return i;
}

         
static char* get(array_t* self, int i) {
  if (self == NULL) { /* array doesn't exist */
    return NULL;
  }

  if (abs(i) > self->_n) {
    return NULL;
  }
  if (i<0) {
    return(self->_p[self->_n+i]);
  } else {
    return(self->_p[i]);
  }
  return NULL; /* should never be hit */
}

static int set(array_t* self, int i, char* s) {
  char* ptr;
  int   err;
  int   index;

  if (self == NULL) { /* array doesn't exist */
    return -1;
  }

  index = i;
  if (i < 0) {
    index = self->_n+i;
  }
  if (index < 0) {
    syslog(LOG_ERR, "out of bounds set call, index of %d of %d items",
           i, self->_n);
    return -1; /* this is an error */
  }
  err = _alloc_array(self, index); /* make sure it's large enough */
  if (err < 0) {
    return -1;
  }
  if (self->_p[index] != NULL) { /* free up old item */
    free(self->_p[index]);
  }
  self->_p[index] = _strdup(s);
  if (index > self->_n) {
    self->_n = index + 1;
  }
  return 0;
}

static int add(array_t* self, char* s) {
  char* ptr;
  int   err;

  if (self == NULL) { /* array doesn't exist */
    return -1;
  }

  self->_n++;
  err = _alloc_array(self, self->_n);
  if (err < 0) {
    return -1;
  }
  self->_p[self->_n-1] = _strdup(s);
  return 0;
}

static void del(array_t* self, int i) {
  int index;
  
  if (self == NULL) { /* array doesn't exist */
    return;
  }

  if (i < 0) {
    index = self->_n + i;
  } else {
    index = i;
  }
  if ((index < 0) || (index >= self->_n)) {
    syslog(LOG_ERR, "out of bounds del call, index of %d of %d items",
           i, self->_n);
    return;
  }
  free(self->_p[index]);
  self->_p[index] = NULL;
  if (index == (self->_n-1)) {
    self->_n--;
  }
  return;
}

static int size(array_t* self) {
  if (self == NULL) { /* array doesn't exist */
    return -1;
  }
  return self->_n;
}

static int setsize(array_t* self, int size) {
  int i;
  int err;
  if (self == NULL) { /* array doesn't exist */
    return -1;
  }
  if (size < self->_n) { /* need to chuck elements */
    for (i=size; i < self->_n; i++) {
      if (self->_p[i] != NULL) {
        free(self->_p[i]);
        self->_p[i] = NULL;
      }
    }
  } else { /* need to add elements */
    err = _alloc_array(self, size);
    if (err < 0) {
      return -1;
    }
  }
  self->_n = size;
  return 0;
}

void
delete_array(array_t* self) {
  int i;
  
  if (self == NULL) {
    return;
  }
  for (i=0; i < self->_n_alloc; i++) {
    if (self->_p[i] != NULL) {
      free(self->_p[i]);
      self->_p[i] = NULL;
    }
  }
  if (self->_p != NULL) {
    free(self->_p);
  }
  free(self);
}

array_t*
new_array(void) {
  array_t* ptr;
  int      err;

  ptr = (array_t*)malloc(sizeof(array_t));
  if (ptr == NULL) {
    syslog(LOG_CRIT, "couldn't allocate memory for array_t : %m");
    return NULL;
  }
  memset((void*)ptr, 0, sizeof(array_t));
  ptr->get = get;
  ptr->set = set;
  ptr->add = add;
  ptr->del = del;
  ptr->size = size;
  ptr->setsize = setsize;
  err = _alloc_array(ptr, 1);
  if (err < 0) {
    free(ptr);
    syslog(LOG_CRIT, "couldn't allocate memory for array storage");
    ptr = NULL;
  }
  return ptr;
}

/* misc array functions */

array_t*
split(char *t, char* s) {
  char*    tmp;
  char*    loc;
  char*    ptr;
  array_t* a;
  int      err;

  if (s == NULL) {
    return NULL;
  }
  a = new_array();
  if (a == NULL) {
    return a;
  }
  tmp = _strdup(s);
  ptr = strtok_r(tmp, t, &loc);
  while (ptr != NULL) {
    err = a->add(a, ptr);
    if (err < 0) {
      syslog(LOG_ERR, "split failed to add [%s] to array", ptr);
      delete_array(a);
      free(tmp);
      return NULL;
    }
    ptr = strtok_r(NULL, t, &loc);
  }
  free(tmp);
  return a;
}
