/*
 * pam_gdbm;  UTC copyright 1999,2000,2001 (UTC - Universal Talkware Corp.)
 * 
 * This implements a pam module subset
 * it currently does authentication
 * it returns PAM_SERVICE_ERROR for all other things
 *
 * This is distributed as is with no warranty expressed or implied.
 * Use of this module is at your own risk and UTC cannot be held
 * liable for any damages, direct or indirect, resulting from
 * the use of this module.
 *
 * The format of the gdbm database is
 * login\0password\0.....
 *
 * originally I was going to use the colon as a separator for the information
 * but the code turned out to be cleaner and faster doing it this way
 *
 * $Id: pam_gdbm.c,v 1.8 2001/03/01 13:30:51 estabroo Exp $
 *
 * to compile this module
 * gcc -fPIC -c pam_gdbm.c
 *
 *  under Solaris link something like this
 * ld -dy -G -o pam_gdbm.so pam_gdbm.o /usr/local/lib/libgdbm.a -lpam
 *
 *  under linux link like this
 * ld -x --shared -o pam_gdbm.so pam_gdbm.o -lpam -lgdbm
 *
 */


/* define the sections we support (or will support) */
/* #define PAM_STATIC */
#define PAM_SM_PASSWORD
#define PAM_SM_ACCOUNT
#define PAM_SM_SESSION
#define PAM_SM_AUTH

/* other defines for locations within the hash and misc */
/* currently only login and password are defined, but anything could be added
   like expiration information (of course more of the module needs to be finished
   before that is functional).  Eventually this should be moved to a .h file
   so that other c programs can use the defines to access the gdbm data file
*/
#define PAM_GDBM_DEFAULT_FILE "/usr/local/etc/users_gdbm"
#define PAM_GDBM_LOGIN 0
#define PAM_GDBM_PASSWORD 1


#include <security/pam_appl.h>
#include <security/pam_modules.h>

#include <unistd.h>
#include <syslog.h>
#include <gdbm.h> /* for me it's here /usr/local/include */

/* handle the EXTERN situation under solaris */
#ifndef PAM_EXTERN
#define PAM_EXTERN
#endif

/* Authentication management section */

PAM_EXTERN int 
pam_sm_authenticate(pam_handle_t* pmh, int flags, int argc, const char** argv) {
  char* user;
  char* password_user;
  char* password_gdbm;
  char  gdbm_file[2048];
  int   gdbm_file_specified;
  int   error;
  int   i;
  int   debug = 0;

  gdbm_file_specified = 0;
  memset((void *)gdbm_file, 0, 2048);
  for (i=0; i < argc; i++) {
    if (!strcmp("debug", argv[i])) {
      debug++; /* verbosity increases with every debug */
    }
    if (!strcmp("gdbm_file", argv[i])) {
      if (i+1 <= argc) {
	i++;
	strncpy(gdbm_file, argv[i], 2047);
	gdbm_file_specified = 1;
      }
    }
  }
  if (gdbm_file_specified == 0) {
    strcpy(gdbm_file, PAM_GDBM_DEFAULT_FILE);
  }
  if (debug) {
    syslog(LOG_DEBUG, "PAM_GDBM: pam_sm_authenticate called");
  }
#ifdef linux
  error = pam_get_user(pmh, (const char**)&user, 0);
#else
  error = pam_get_user(pmh, &user, 0);
#endif
  if (error != PAM_SUCCESS) {
    syslog(LOG_ERR, "%s", pam_strerror(pmh, error));
    return PAM_AUTHINFO_UNAVAIL;
  }
  
  /* this block can return, PAM_AUTHINFO_UNAVAIL or PAM_USER_UNKNOWN */
  error = pam_gdbm_get_item(gdbm_file, user, PAM_GDBM_PASSWORD,
			    &password_gdbm, debug);
  if (error != PAM_SUCCESS) {
    syslog(LOG_ERR, "PAM_GDBM: %s", pam_strerror(pmh, error));
    return error;
  }

  /* check for allowance of null password */
  if (password_gdbm == 0) {
    syslog(LOG_ALERT, "PAM_GDBM: null password found for [%s]", user);
    if (flags & PAM_DISALLOW_NULL_AUTHTOK) {
      syslog(LOG_ALERT, "PAM_GDBM: null password not allowed [%s]", user);
      return PAM_AUTH_ERR;
    } else {
      syslog(LOG_ALERT, "PAM_GDBM: null password allowed for [%s]", user);
      return PAM_SUCCESS;
    }
  }

  password_user = 0;
  error = pam_gdbm_get_password(pmh, &password_user, debug);
  if ((error != PAM_SUCCESS) || (password_user == 0)) {
    syslog(LOG_ERR, "PAM_GDBM: %s", pam_strerror(pmh, error));
    return PAM_AUTHINFO_UNAVAIL;
  }
  error = pam_gdbm_compare_password(password_user, password_gdbm, debug);
  free(password_gdbm);
  return error;
}

PAM_EXTERN int 
pam_sm_setcred(pam_handle_t* pmh, int flags, int argc, const char** argv) {
  int i;
  int debug = 0;

  for (i=0; i < argc; i++) {
    if (!strcmp("debug", argv[i])) {
      debug++; /* verbosity increases with every debug */
    }
  }
  if (debug) {
    syslog(LOG_DEBUG, "PAM_GDBM: pam_sm_setcred called");
  }
  return PAM_SERVICE_ERR;
}



/* PAM account management section */

PAM_EXTERN int 
pam_sm_acct_mgmt(pam_handle_t* pmh, int flags, int argc, const char** argv) {
  int i;
  int debug = 0;

  for (i=0; i < argc; i++) {
    if (!strcmp("debug", argv[i])) {
      debug++; /* verbosity increases with every debug */
    }
  }
  if (debug) {
    syslog(LOG_DEBUG, "PAM_GDBM: pam_sm_acct_mgmt called");
  }
  return PAM_SUCCESS;
}

/* PAM session management section */

PAM_EXTERN int 
pam_sm_open_session(pam_handle_t* pmh, int flags, int argc, const char** argv) {
  int i;
  int debug = 0;

  for (i=0; i < argc; i++) {
    if (!strcmp("debug", argv[i])) {
      debug++; /* verbosity increases with every debug */
    }
  }
  if (debug) {
    syslog(LOG_DEBUG, "PAM_GDBM: pam_sm_open_session called");
  }
  return PAM_SERVICE_ERR;
}

PAM_EXTERN int 
pam_sm_close_session(pam_handle_t* pmh, int flags, int argc, const char** argv) {
  int i;
  int debug = 0;

  for (i=0; i < argc; i++) {
    if (!strcmp("debug", argv[i])) {
      debug++; /* verbosity increases with every debug */
    }
  }
  if (debug) {
    syslog(LOG_DEBUG, "PAM_GDBM: pam_sm_close_session called");
  }
  return PAM_SERVICE_ERR;
}



/* PAM password management */

PAM_EXTERN int 
pam_sm_chauthtok(pam_handle_t* pmh, int flags, int argc, const char** argv) {
  int i;
  int debug = 0;

  for (i=0; i < argc; i++) {
    if (!strcmp("debug", argv[i])) {
      debug++; /* verbosity increases with every debug */
    }
  }
  if (debug) {
    syslog(LOG_DEBUG, "PAM_GDBM: pam_sm_chauthtok called");
  }
  return PAM_PERM_DENIED; /* doesn't do anything yet */
}


/* GDBM access functions */
PAM_EXTERN int
pam_gdbm_get_item(char* gdbm_file, char* user, int item, char** item_ptr, int debug) {
  GDBM_FILE my_db;
  datum     query;
  datum     ans;
  char*     ans_ptr;
  char*     ptr;
  int       size;
  int       i;
  int       item_n;

  if (debug) {
    syslog(LOG_DEBUG, "PAM_GDBM: pam_gdbm_get_item called");
  }
  /* scan user string to make sure we don't core or anything */
  my_db = gdbm_open(gdbm_file, 512, GDBM_READER, 0444, 0);
  if (my_db == 0) {
    syslog(LOG_ALERT, "PAM_GDBM: couldn't open gdbm file [%s]", gdbm_file);
    return PAM_AUTHINFO_UNAVAIL;
  }
  query.dptr = (void *)user;
  query.dsize = strlen(user);

  if (debug > 1) {
    syslog(LOG_DEBUG, "PAM_GDBM: looking for [%s] in [%s]", query.dptr, gdbm_file);
  }
  ans = gdbm_fetch(my_db, query);
  if (ans.dptr == (char *)0) {
    gdbm_close(my_db);
    syslog(LOG_INFO, "PAM_GDBM: user not found [%s]", user);
    return PAM_USER_UNKNOWN;
  }
  for (i=0, item_n=0, ans_ptr = ans.dptr; (i < ans.dsize) && (item_n < item);
       i++, ans_ptr++) {
    if (ans_ptr[0] == 0) {
      item_n++;
    }
  }
  if (i >= ans.dsize) { /* item not found */
    gdbm_close(my_db);
    if (debug) {
      syslog(LOG_DEBUG, "PAM_GDBM: item [%d] not found for user [%s]", item, user);
    }
    *item_ptr = 0;
    return PAM_AUTHINFO_UNAVAIL;
  }
  size = strlen(ans_ptr)+1;
  if (debug > 2) {
    syslog(LOG_DEBUG, "PAM_GDBM: item [%s] is [%d] bytes", ans_ptr, size);
  }
  ptr = (char *)malloc(sizeof(char *)*size);
  if (ptr == (char *)0) {
    syslog(LOG_CRIT, "PAM_GDBM: couldn't malloc space for item");
    gdbm_close(my_db);
    *item_ptr = 0;
    return PAM_AUTHINFO_UNAVAIL;
  }
  memset((void *)ptr, 0, size);
  strncpy(ptr, ans_ptr, size-1);
  gdbm_close(my_db);
  *item_ptr = ptr;
  return PAM_SUCCESS;
}

PAM_EXTERN int
pam_gdbm_compare_password(char* password_user, char* password_gdbm, int debug) {
  char* crypted_ptr;

  if (debug) {
    syslog(LOG_DEBUG, "PAM_GDBM: pam_gdbm_compare_password called");
  }
  crypted_ptr = (char *)crypt(password_user, password_gdbm);
  if (!strcmp(crypted_ptr, password_gdbm)) {
    if (debug > 1) {
      syslog(LOG_DEBUG, "PAM_GDBM: passwords matched");
    }
    return PAM_SUCCESS;
  }
  if (debug > 1) {
    syslog(LOG_DEBUG, "PAM_GDBM: password match failed");
  }
  return PAM_AUTH_ERR;
}

PAM_EXTERN int
pam_gdbm_get_password(pam_handle_t* pmh, char** password_user, int debug) {
  int                  error;
  struct pam_message   message;
  struct pam_message*  mesg_ptr;
  struct pam_response* response;
  void*                ptr;
  struct pam_conv*     conv_ptr;

  if (debug) {
    syslog(LOG_DEBUG, "PAM_GDBM: pam_gdbm_get_password called");
  }
  mesg_ptr = &message;
#ifdef linux
  error = pam_get_item(pmh, PAM_AUTHTOK, (const void **)(password_user));
#else
  error = pam_get_item(pmh, PAM_AUTHTOK, (void **)(password_user));
#endif
  if (error != PAM_SUCCESS) {
    return error;
  }
  /* if authtoken is not yet here use conv routine to prompt for it */
  if (*password_user == (char *)0) {
    message.msg_style = PAM_PROMPT_ECHO_OFF;
    message.msg = 0;
#ifdef linux
    error = pam_get_item(pmh, PAM_CONV, (const void **)&ptr);
#else
    error = pam_get_item(pmh, PAM_CONV, &ptr);
#endif
    if (error != PAM_SUCCESS) {
      return error;
    }
    conv_ptr = (struct pam_conv*)ptr;
    if (conv_ptr == (struct pam_conv*)0) {
      return PAM_AUTHINFO_UNAVAIL;
    }
#ifdef linux
    error = conv_ptr->conv(1, (const struct pam_message **)&(mesg_ptr), &response, 
			   conv_ptr->appdata_ptr);
#else
    error = conv_ptr->conv(1, &(mesg_ptr), &response, conv_ptr->appdata_ptr);
#endif
    if (error != PAM_SUCCESS) {
      return error;
    }
    /* record auth token for later use */
#ifdef linux
    error = pam_set_item(pmh, PAM_AUTHTOK, (const void *)(response->resp));
#else
    error = pam_set_item(pmh, PAM_AUTHTOK, (void *)(response->resp));
#endif
    if (error != PAM_SUCCESS) {
      free(response->resp);
      free(response);
      return error;
    }
    free(response->resp);
    free(response);
    
  }
#ifdef linux
  error = pam_get_item(pmh, PAM_AUTHTOK, (const void **)(password_user));
#else
  error = pam_get_item(pmh, PAM_AUTHTOK, (void **)(password_user));
#endif
  return error;
}

