/* 
 * yaps - yet another process status reporter
 * prints out process information kind of like ps
 * allows a process to 'rename' itself
 * yaps runs under Solaris
 * compile with gcc -O2 -o yaps yaps.c
 * then chmod 4555 so it is suid root
 *
 * $Id: yaps.c,v 1.10 1999/04/11 23:50:47 eric Exp $
 *
 * copyright 1999 Eric Estabrooks
 * This program comes as is with no warranty expressed or implied
 * I am not responsible for any damages that may result from the use
 * of this program.  Use at your own risk.
 * 
 * If you make changes to the program, please note the changes before
 * redistribution.
 *
 * submit patches or suggestions to eric@urbanrage.com
 */

#include <sys/types.h>
#include <sys/signal.h>
#include <sys/fault.h>
#include <sys/syscall.h>
#include <sys/procfs.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <dirent.h>
#include <pwd.h>
#include <time.h>
#include <limits.h>
#include <sys/sysmacros.h>
#include <search.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

/* we use the gnu version */
#include "getopt.h"

/* we are going to use /tmp/ps_data to determine term info */
#define PS_DATA_STRING_SIZE 16
#define PS_INFO_ITEMS 12
#define N_LONG_OPTIONS

int open_ps_data(void);

typedef struct {
  char name[16];
  long dev;
} data_t;

typedef struct {
  int  order;
  char hashkey[16];
  char name[16];
  char name_fmt[16];
  char data_fmt[16];
  int  data_type; /* 0 = char, 1 = int, 2 = byte */
  char **data;
} info_t;

typedef struct {
  int order[PS_INFO_ITEMS];
  int n_items;
} info_list_t;

extern char *optarg;

char cvs_vers[] = "$Id: yaps.c,v 1.10 1999/04/11 23:50:47 eric Exp $";

void
usage(char *p_name) {
  printf("\nHelp for yaps version %s\n\n", _YAPS_VERS);
  printf("usage: %s [-leh] [-o format] [--help] [--wide] [--grep string] [--grepv string]\n",
	 p_name);
  printf("\t-h, --help   print this information\n");
  printf("\t-l           long format prints s, uid, pid, ppid, nice, time, tty, cmd\n");
  printf("\t-e           prints information about all processes\n");
  printf("\t--wide       prints pid, ppid, uid, stime, tty, time, ccmd, cmd\n");
  printf("\t--fgrep      only prints processes whose cmd contains the string\n");
  printf("\t--fgrepv     only prints processes whose cmd doesn't contain the string\n");
  printf("\t              grep and grepv compare against cmd and ccmd, this may\n");
  printf("\t              proceses to be printed that do/don't have the string\n");
  printf("\t-o           print whatever information you want (that's available)\n");
  printf("\t              the format string is a comma delimited string\n");
  printf("\t              format must contain one or more of:\n");
  printf("\t                 pid ppid uid stime tty time cmd s ccmd nice ruid euid\n");
  exit(0);
}

int
main (int argc, char *argv[]) { /* currently you must specify pids */
  int             retval;
  int             fildes;
  int             i,j;
  int             n_bytes;
  int             n_items;
  int             longind; /* for getopts_long */
  int             show_all_uid = 0;
  gid_t           who_i_am;
  double          minutes;
  double          seconds;
  const int       proc_name_length = 80;
  unsigned long   proc_name_offset;
  off_t           offset;
  struct prpsinfo psinfo;
  unsigned char   dir_buff[8192];
  char            path[8192];
  char            proc_time[16];
  char            proc_name[4096];
  char            ruid_buff[1024];
  char            euid_buff[1024];
  char            stime_buff[27];
  char            key_buff[16];
  char            o_buff[1024]; /* used for the -o option */
  char            *cmd;
  char            *ccmd = psinfo.pr_psargs;
  char            *ruid = ruid_buff;
  char            *euid = euid_buff;
  char            *stime = stime_buff;
  char            *tty_ptr;
  char            *time_ptr = proc_time;
  char            *nice_ptr = (char *)&psinfo.pr_nice;
  char            *state_ptr = (char *)&psinfo.pr_sname;
  char            *grep_val = 0;
  char            *grepv_val = 0;
  char            *token_ptr = 0;
  char            *o_ptr = 0; /* next item to add to info_list */
  DIR             *proc_dir;
  ENTRY           item;
  ENTRY           *item_ptr;
  time_t          now;
  struct dirent   *ent_ptr;
  struct passwd   pwd;
  struct passwd   *pwd_ptr;
  data_t          *data_storage;
  data_t          *data_ptr;
  struct option   long_options[] = {
    { "fgrep",  1, 0, 'Z' },
    { "fgrepv", 1, 0, 'Y' },
    { "wide",  0, 0, 'W' },
    { "help",  0, 0, 'h' },
    { 0,0,0,0 },
  };
  info_list_t     o_list[] = { { { 0 }, 0 } };
  info_list_t     w_list[] = { { { 0,1,2,3,4,5,11,6 }, 8 } };
  info_list_t     default_list[] = { { { 0,1,2,3,4,5,6 }, 7 } };
  info_list_t     long_list[] = { { { 10,2,0,1,7,4,5,6 }, 8 } };
  info_list_t     *info_order = default_list;
  info_t          info[] = { 
    { 0,  "pid",   "PID",   "%5s",   "%5d",    1, 0 },
    { 1,  "ppid",  "PPID",  "%5s",   "%5d",    1, 0 },
    { 2,  "uid",   "UID",   "%8s",   "%8s",    0, 0 },
    { 3,  "stime", "STIME", "%8s",   "%8s",    0, 0 },
    { 4,  "tty",   "TTY",   "%-8s",  "%-8s",   0, 0 },
    { 5,  "time",  "TIME",  "%8s",   "%8s",    0, 0 },
    { 6,  "cmd",   "CMD",   "%s",    "%s",     0, 0 },
    { 7,  "nice",  "NI",    "%2s",   "%2d",    2, 0 },
    { 8,  "euid",  "EUID",  "%8s",   "%8s",    0, 0 },
    { 9,  "ruid",  "RUID",  "%8s",   "%8s",    0, 0 },
    { 10, "s",     "S",     "%1s",   "%c",     2, 0 },
    { 11, "ccmd",  "CCMD",  "%-60s", "%-60s",  0, 0 },
  };

  /* initialize info_t data portion hopefully this will appease sunpro c */
  info[0].data = (char **)&psinfo.pr_pid;
  info[1].data = (char **)&psinfo.pr_ppid;
  info[2].data = &euid;
  info[3].data = &stime;
  info[4].data = &tty_ptr;
  info[5].data = &time_ptr;
  info[6].data = &cmd;
  info[7].data = &nice_ptr;
  info[8].data = &euid;
  info[9].data = &ruid;
  info[10].data = &state_ptr;
  info[11].data = &ccmd;

  /* put info items in hash for -o ordering */
  hcreate(512);
  for (i=0; i < PS_INFO_ITEMS; i++) {
    item.key = info[i].hashkey;
    item.data = (char *)info[i].order;
    hsearch(item, ENTER);
  }
  
  /* this is who I really am */
  who_i_am = getuid();

  /* handle args */
  for (retval = getopt_long(argc, argv, "leho:", long_options, &longind); retval != -1;
       retval = getopt_long(argc, argv, "leho:", long_options, &longind)) {
    switch (retval) {
    case 'Z' : grep_val = optarg; break;
    case 'Y' : grepv_val = optarg; break;
    case 'W' : info_order = w_list; break;
    case 'l' : info_order = long_list; break;
    case 'e' : show_all_uid = 1; break;
    case 'h' : usage(argv[0]); break;
    case 'o' : {
      memset((void *)o_buff, 0, 1024);
      strncpy(o_buff, optarg, 1023); /* copy but don't overflow */
      for (i=0, o_ptr = (char *)strtok_r(o_buff, ",", &token_ptr); o_ptr != 0;
	   o_ptr = (char *)strtok_r(0, ",", &token_ptr), i++) {
	item.key = o_ptr;
	item_ptr = hsearch(item, FIND);
	if (item_ptr == (ENTRY *)0) {
	  fprintf(stderr, "illegal -o option [%s]\n", o_ptr);
	  exit(0);
	}
	o_list[0].order[i] = (int)(item_ptr->data);
	o_list[0].n_items = i+1;
      }
      if (i != 0) {
	info_order = o_list;
      } else {
	fprintf(stderr, "no options given for -o flag\n");
	exit(0);
      }
    }; break;
    default : {
      fprintf(stderr, "unknown flag value %c\n", retval);
      exit(0);
    }
    }; /* end of arg switch */
  }
  if (who_i_am == 0) {
    show_all_uid = 1;
  }

  /* read in process info */
  fildes = open_ps_data();
  if (fildes < 0) {
    fprintf(stderr,"Couldn't open /tmp/ps_data; please do a ps -aef\n");
    exit(errno);
  }
  n_bytes = read(fildes, dir_buff, 4);
  n_items = *((int *)dir_buff);
  /* printf("number of items = %d\n",n_items);*/
  data_storage = (data_t *)calloc(n_items, sizeof(data_t));
  for (i=0; i < n_items; i++) {
    data_ptr = &data_storage[i];
    n_bytes = read(fildes, (void *)data_ptr,sizeof(data_t));
    if (n_bytes == sizeof(data_t)) {
      sprintf(key_buff, "%ld", data_ptr->dev);
      item.key = key_buff;
      item.data = data_ptr->name;
      item_ptr = hsearch(item, ENTER);
      /*      printf("added [%s] [%s] [%x]\n", item.key, data_ptr->name, item_ptr); */
    } else {
      printf("failed at %d\n", i);
      break;
    }
  }
  close(fildes);

  proc_dir = opendir("/proc");
  if (proc_dir == (DIR *)0) {
    perror("Couldn't open /proc");
    exit(errno);
  }
  /* construct the info header */
  printf(info[info_order->order[0]].name_fmt, info[info_order->order[0]].name);
  for (i=1; i < info_order->n_items; i++) {
    printf(" ");
    printf(info[info_order->order[i]].name_fmt, info[info_order->order[i]].name);
  }
  printf("\n");
#ifdef _POSIX_PTHREAD_SEMANTICS
  for (retval = readdir_r(proc_dir, (void *)dir_buff, &ent_ptr); retval == 0;
       retval = readdir_r(proc_dir, (void *)dir_buff, &ent_ptr))
#else
  for (ent_ptr = (struct dirent *)readdir_r(proc_dir, (void *)dir_buff); ent_ptr != 0;
       ent_ptr = (struct dirent *)readdir_r(proc_dir, (void *)dir_buff))
#endif
    {
      if ((strcmp(ent_ptr->d_name, ".") == 0) ||
	  (strcmp(ent_ptr->d_name, "..") == 0)) {
	continue;
      }
      sprintf(path, "/proc/%s", ent_ptr->d_name);
      fildes = open(path, O_RDONLY);
      if (fildes < 0) {
	/*perror(path);*/
	continue;
      }
      memset((void *)&psinfo, 0, sizeof(psinfo));
      retval = ioctl(fildes, PIOCPSINFO, &psinfo);
      if (retval < 0) {
	printf("PIOCPSINFO failed on %s", ent_ptr->d_name);
	close(fildes);
	continue;
      }

      /* check uid */
      if (show_all_uid == 0) {
	if (psinfo.pr_uid != who_i_am) {
	  close(fildes);
	  continue;
	}
      }

      /* get process name information */
      psinfo.pr_psargs[60] = 0; /* limit size of ccmd print */
      if (psinfo.pr_zomb != 0) {
	sprintf(proc_name, "<defunct>");
	cmd = proc_name;
      } else {
	offset = lseek(fildes, (off_t)psinfo.pr_argv, SEEK_SET);
	if (offset == -1) {
	  close(fildes);
	  continue;
	}
	n_bytes = read(fildes, (void *)&proc_name_offset, sizeof(unsigned long));
	offset = lseek(fildes, (off_t)proc_name_offset, SEEK_SET);
	if (offset == -1) {
	  perror("lseek failed");
	  close(fildes);
	  continue;
	}
	memset((void *)proc_name, 0, 4096);
	n_bytes = read(fildes, (void *)proc_name, proc_name_length);
	if (n_bytes <= 0) {
	  cmd = psinfo.pr_psargs;
	} else {
	  cmd = proc_name;
	  /* clean up proc_name */
	  for (j=0; j < n_bytes; j++) {
	    if (!isprint((int)proc_name[j])) {
	      proc_name[j] = 0;
	    }
	  }
	}
      }
      
      /* do we output info on this process */
      if (grep_val != 0) {
	if ((strstr(cmd, grep_val) == 0) && (strstr(ccmd, grep_val) == 0)) {
	  /* skip this one, it doesn't contain the string */
	  close(fildes);
	  continue;
	}
      }
      if (grepv_val != 0) {
	if ((strstr(cmd, grepv_val) != 0) || (strstr(ccmd, grepv_val) != 0)) {
	  /* skip this one, it contains the string */
	  close(fildes);
	  continue;
	}
      }

      /* generate ruid & euid*/
      pwd_ptr = getpwuid_r(psinfo.pr_uid, &pwd, ruid_buff, 1024);
      if (pwd_ptr != 0) {
	ruid = pwd.pw_name;
      } else {
	sprintf(ruid_buff,"%16ld", psinfo.pr_uid);
	ruid = ruid_buff;
      }
      pwd_ptr = getpwuid_r(psinfo.pr_euid, &pwd, euid_buff, 1024);
      if (pwd_ptr != 0) {
	euid = pwd.pw_name;
      } else {
	sprintf(euid_buff,"%16ld", psinfo.pr_uid);
	euid = ruid_buff;
      }

      /* generate tty information */
      if (psinfo.pr_ttydev == PRNODEV) {
	if (psinfo.pr_zomb != 0) {
	  tty_ptr = " ";
	} else {
	  tty_ptr = "?";
	}
      } else {
	sprintf(key_buff, "%ld", psinfo.pr_ttydev);
	item.key = key_buff;
	item_ptr = hsearch(item, FIND);
	if (item_ptr == (ENTRY *)0) {
	  tty_ptr = "?";
	} else {
	  if (item_ptr->data == (char *) 0) {
	    tty_ptr = "bad hash";
	  } else {
	    tty_ptr = item_ptr->data;
	  }
	}
      }
      /* generate start time */
      if (psinfo.pr_zomb != 0) {
	sprintf(stime_buff, " ");
      } else {
	now = time(0);
	if ((now - psinfo.pr_start.tv_sec) > 86399) {
	  retval = cftime(stime_buff, "%b %d", &psinfo.pr_start.tv_sec);
	} else {
	  retval = cftime(stime_buff, "%H:%M:%S", &psinfo.pr_start.tv_sec);
	}
      }

      /* generate cpu time */
      minutes = (psinfo.pr_time.tv_sec + psinfo.pr_time.tv_nsec / 1000000000.0 + 
		 0.4999) / 60;
      seconds = (minutes - (long)minutes) * 60;
      sprintf(proc_time, "%0ld:%02ld", (long)(minutes), (long)(seconds));

      /* dump the information */
      for(i=0; i < info_order->n_items; i++) {
	if (i != 0) {
	  printf(" ");
	}
	switch (info[info_order->order[i]].data_type) {
	case 0 :
	  printf(info[info_order->order[i]].data_fmt, 
		 *(info[info_order->order[i]].data));
	  break;
	case 1 :
	  printf(info[info_order->order[i]].data_fmt, 
		 (int)(*(info[info_order->order[i]].data)));
	  break;
	case 2 :
	  printf(info[info_order->order[i]].data_fmt, 
		 (char)(**(info[info_order->order[i]].data)) & 0xff);
	  break;
	default :
	  printf("unknown data type\n");
	};
      }
      printf("\n");
      close(fildes);
    }
  closedir(proc_dir);
  return(0);
}

int
open_ps_data() {
  int  fildes;
  int  pipedes[2];
  int  pid;
  int  n_bytes = 1024;
  char buff[1024];

  fildes = open("/tmp/ps_data", O_RDONLY);
  if (fildes < 0) {
    if (errno == ENOENT) {
      pipe(pipedes);
      pid = fork();
      if (pid < 0) {
	return(-1);
      }
      if (pid == 0) {
	close(fileno(stdout));
	close(fileno(stderr));
	dup(pipedes[1]);
	dup(pipedes[1]);
	execl("/bin/ps", 0); /* cheap easy way of building /tmp/ps_data */
      }
      close(pipedes[1]);
      while (n_bytes == 1024) {
	n_bytes = read(pipedes[0], buff, 1024);
      }
      fildes = open("/tmp/ps_data", O_RDONLY);
      return fildes;
    } else {
      return(-1);
    }
  }
  return fildes;
}
